/***************************************************************************
                                  BasicCoach.C 
                             -------------------
    begin                : 2001
    copyright            : (C)2001-2003 Multi-Agent Systems Laboratory,
                           University of Science and Technology of China 
                           Zhanxiang Huang, Yang Yang and Xiaoping Chen.
    email                : huangzx@mail.ustc.edu.cn
 ***************************************************************************/

#include "BasicCoach.h"

using namespace std;

BasicCoach::BasicCoach()
{
        sock = NULL;
        m_parse = NULL;
        bMsgCome = false;
        bSeeMsgCome = false;
        m_tmpWMS.SetAllDataInvalidate();
}

BasicCoach::BasicCoach(UDPSocket *s, Parse *parse , string &ourTeamName)
{
       SetConnection(s);
       SetParsePtr(parse);
       wm.SetOurTeamName(ourTeamName);
}


bool BasicCoach::SetConnection(UDPSocket *s)
{
        sock = s;
        return true;
}

bool BasicCoach::SetParsePtr(Parse *parse)
{
        m_parse = parse;
        return true;
}




// say/send msg to the server
int BasicCoach::Send(const string &msg) const
{
    if(msg=="") return -1;
    return sock->SendMsg(msg);
}

int BasicCoach::Say(const string &msg) const
{
   string buf;
   
   if(sock==NULL){
       return -1;  // socket have not been initialized
   }
   
   buf = "(say " + msg + ")";
   
   return Send(buf);
}

// function that execute every cycle
int BasicCoach::UpdateRoutine()
{
    // coach ̣ msg ģʽ
    //     coach UpdateroutineÿʱǷµϢûµϢ
    // һʱ䣬鿴Ƿsee msg,Ϣģִͣв
    string msg;
    list<string>::iterator  iter;
    MsgT    msgT;
    MsgState msgs = MsgS_No_Msg_Come;
    
    //check if there are new msgs
    if(!bMsgCome){
        //ûϢ
        return msgs;
    }else{
        msgs = MsgS_Msg_Come;
    }
     
    // deal msg that received
    for( iter=msgQ.begin(); iter!=msgQ.end(); iter++){
        msg = *iter;
        msgT = m_parse->ParseMsg(msg, m_tmpWMS);
        
        if(msgT == MSG_SEE){
            // յsee msg
            //bSeeMsgCome = true;
            msgs = MsgS_See_Msg_Come;

            UpdateWorldModel(m_tmpWMS);
       
            // when Update current WorldModel
            // re-initialize the m_tmpWMS, wait for next Update
            m_tmpWMS.SetAllDataInvalidate();
    
            // play_mode,score,timeЧ,wait for next Update
            m_tmpWMS.SetTimeStamp( wm.GetCurrentTime() );
            m_tmpWMS.SetPlayMode( wm.GetPlayMode( wm.GetCurrentTime() ) );
            m_tmpWMS.SetOurScore( wm.GetOurScore( wm.GetCurrentTime() ) );
            m_tmpWMS.SetOppScore( wm.GetOppScore( wm.GetCurrentTime() ) );
        }
    }

    // clear the messge queue that have been dealed with
    msgQ.clear();
    // Update World Model only when recv see msg

    // thinking
    //Thinking();
  
    bMsgCome = false;
    
    return msgs;    
}


//  send msgs received to coach, msgs will be added to msg queue
int BasicCoach::AddNewMsg(string &msg)
{
    bMsgCome = true;
    msgQ.push_back(msg);
    return 0;
}


//  Update World Model
int BasicCoach::UpdateWorldModel(WMSnapShot &tmpWMS)
{
         list<EventT>  events;
         list<EventT>::iterator  iter;
         WMSnapShot  wms;
         
         int iVal;
         int tm;
         char buf[64];  
         
        // ¼
        events = tmpWMS.GetEvents();
        for( iter = events.begin(); iter != events.end(); iter++){
                switch( (*iter).name ){
                        case EV_SEE_OUR_GOALIE:
                                if( !wm.IsOurGoalieNumSet() ){
                                        // Աû
                                        strcpy(buf, (*iter).value.c_str());
                                        sscanf(buf,"%d", &iVal);
                                        wm.SetOurGoalieNum(iVal);
                                }
                                //events.erase(iter);
                                break;
                        case EV_SEE_OPP_GOALIE:
                                if( !wm.IsOppGoalieNumSet() ){
                                        strcpy(buf, (*iter).value.c_str());
                                        sscanf(buf,"%d", &iVal);
                                        wm.SetOppGoalieNum(iVal);
                                }
                                //events.erase(iter);
                                break;
                        case EV_SEE_OPP_NAME:
                                if( !wm.IsOppTeamNameSet() ){
                                        wm.SetOppTeamName( (*iter).value );
                                }
                                //events.erase(iter);
                                break;
                        case EV_MISS_SEE_MSG:
                                // ȱ
                                sscanf( (*iter).value.c_str(), "%d", &iVal);
                                // һWMSnapShotΪֵWorldModel
                                wms = wm.GetWMSnapShot( wm.GetCurrentTime() );
                                for(int i=0;i<iVal;i++){
                                    // wmstimeStamp
                                    wms.SetTimeStamp( wms.GetTimeStamp()+i+1 ) ;
                                    wm.UpdateWorldModel(wms);
                                }
                                //events.erase(iter);
                                break;
                        case EV_GOAL_LEFT:
                                // read the goal
                                sscanf( (*iter).value.c_str(),"%d %d", &iVal , &tm);
                                if(wm.GetOurSide() == FIELD_SIDE_L)
                                    tmpWMS.SetOurScore(iVal);
                                else
                                    tmpWMS.SetOppScore(iVal);
                                //events.erase(iter);
                                break;
                        case EV_GOAL_RIGHT:
                                // read the goal
                                sscanf( (*iter).value.c_str(),"%d %d", &iVal , &tm);
                                if(wm.GetOurSide() == FIELD_SIDE_R)
                                    tmpWMS.SetOurScore(iVal);
                                else
                                    tmpWMS.SetOppScore(iVal);
                                //events.erase(iter);
                                break;
                        case EV_GOAL_KICK_LEFT:
                              // goal kick for left team 
                              break;
                        case EV_GOAL_KICK_RIGHT:
                              // goal kick for right team 
                              break;
                        case EV_OFFSIDE_LEFT:
                             // offside for left team
                             break;
                        case EV_OFFSIDE_RIGHT:
                            // offside for right team
                            break;
                        case EV_FREE_KICK_LEFT:
                            // free kick for left team
                            break;
                        case EV_FREE_KICK_RIGHT:
                            // free kick for right team
                            break;
                        case EV_KICK_OFF_LEFT:
                            // kick off for left team   
                            break;
                        case EV_KICK_OFF_RIGHT:
                            // kick off for right team  
                            break;
                        case EV_KICK_IN_LEFT:
                            // kick in for left team   
                            break;
                        case EV_KICK_IN_RIGHT:
                            // kick in for right team
                            break;
                        case EV_CORNER_KICK_LEFT:
                            //corner kick left team   
                            break;
                        case EV_CORNER_KICK_RIGHT:
                            //corner kick right team  
                            break;
                        case EV_DROP_BALL:
                            //drop
                            break;
                        default:;
               };
             
        } // end for
        
        tmpWMS.clearEvent();

        // wmsģ

        wm.UpdateWorldModel(tmpWMS);
        
        return 0 ;
}


WorldModel* BasicCoach::GetWorldModel()
{
    return &wm;
}


