/***************************************************************************
                                  Parse.C 
                             -------------------
    begin                : 2001
    copyright            : (C)2001-2003 Multi-Agent Systems Laboratory,
                           University of Science and Technology of China 
                           Zhanxiang Huang, Yang Yang and Xiaoping Chen.
    email                : huangzx@mail.ustc.edu.cn
 ***************************************************************************/

#include "Parse.h"


const  int  DEL = 0;    // delimeter
const  int  LP  =  1;
const  int  RP  =  2;
const  int  STR = 3;
const  int  WORD = 4;

Parse::Parse(const string &teamName)
{
//    Parse(teamName.c_str());    construct function cannot be called!!!!
    SetOurTeamName(teamName.c_str());

    bOppNameSet = false;
    bOurNameSet = true;
        
    bOppGoalieSet = false;
    bOurGoalieSet = false;
}

Parse::Parse(const char *teamName)
{
    SetOurTeamName(teamName);

    bOppNameSet = false;
    bOurNameSet = true;
        
    bOppGoalieSet = false;
    bOurGoalieSet = false;
}


//Function which parse msgs from server and Set correspondent informations to
MsgT Parse::ParseMsg(const string &Msg,WMSnapShot &wms)
{
    return ParseMsg(Msg.c_str(),wms);
}

MsgT Parse::ParseMsg(const char* Msg,WMSnapShot &wms)
{
        char *pszMsg = (char*)Msg;

        switch(pszMsg[1]){
        case 'i':
                //parse initial message : (init l|r ok)
                return MSG_INIT;
                break;
        case 'p':
                if(pszMsg[8]=='t'){
                        // player_type msg
                        if ( ParsePlayerType(Msg, wms) == -1){
                                return MSG_UNEXPECTED;
                        }
                        return MSG_PLAYER_TYPE;
                }
                else{
                        // player_param msg
                        if ( ParsePlayerParam(Msg,wms) == -1){
                                return MSG_UNEXPECTED;
                        }
                        return MSG_PLAYER_PARAM;
                }
        case 's':
                if(pszMsg[3]=='r'){
                        // server_param msg
                        if ( ParseServerParam(Msg,wms) == -1){
                                return MSG_UNEXPECTED;
                        }
                        return MSG_SERVER_PARAM;
                }
                else{
                        // see_global msg
                        // skip "(see_global "
                        //cout<<pszMsg<<endl;
                        if ( ParseSee(pszMsg+11,wms) == -1){
                                return MSG_UNEXPECTED;
                        }
                        return MSG_SEE;
                }
        case 'h':
                // hear msg
                // skip "(hear "
                if ( ParseHear(pszMsg+5,wms) == -1){
                        return MSG_UNEXPECTED; 
                }
                return MSG_HEAR;
        case 'e':
                // error msg
                if ( ParseError(pszMsg,wms) == -1){
                        return MSG_UNEXPECTED;
                }
                return MSG_ERROR;
        case 'o':
                // ok ... msg
                if ( ParseACK(pszMsg,wms) == -1){
                        return MSG_UNEXPECTED;
                }
                return MSG_ACK;
        case 'w':
                // warning msg
                if ( ParseWarning(pszMsg,wms) == -1){
                        return MSG_UNEXPECTED;
                }
                return MSG_WARNING;
        default:
                return MSG_UNEXPECTED;
        };
}


int Parse::ParseHear(const char* Msg,WMSnapShot &wms)
{
        int tm;
        char buf[64];
        int temp;
        char tempbuf[3];
        EventT event;
        char *msg = (char *)Msg;
        
        GetToken(msg,buf);  // Get time
        sscanf(buf,"%d",&tm);

        // ϢʱСڵǰʱ,old msg
        if( tm < wms.GetTimeStamp() ){
            return -1;  // don't do anything
        }   
        
        GetToken(msg,buf); // Get next token

        if(buf[0] == 'r'){
                // referee sound
                GetToken(msg,buf); // Get next token
                switch(buf[0]){
                case 'b':
                        return 0;
                case 't':
                        // time over
                        wms.SetPlayMode(PM_TIME_OVER);
                        return 0;
                case 'p':  // play on
                        wms.SetPlayMode(PM_PLAY_ON);
                        return 0;
                case 'k':  
                        if(buf[5]=='o'){
                                // kick off
                                if(buf[9]=='l'){
                                        wms.SetPlayMode(PM_KICK_OFF_LEFT);
                                        // ¼¼
                                        event.name = EV_KICK_OFF_LEFT;
                                        wms.AddEvent(event);
                                }
                                else{
                                        wms.SetPlayMode(PM_KICK_OFF_RIGHT);
                                        // ¼¼
                                        event.name = EV_KICK_OFF_RIGHT;
                                        wms.AddEvent(event);
                                }
                        }
                        else{
                                // kick in 
                                if(buf[8]=='l'){
                                        wms.SetPlayMode(PM_KICK_IN_LEFT);
                                        // ¼¼
                                        event.name = EV_KICK_IN_LEFT;
                                        wms.AddEvent(event);
                                }
                                else{
                                        wms.SetPlayMode(PM_KICK_IN_RIGHT);
                                        // ¼¼
                                        event.name = EV_KICK_IN_RIGHT;
                                        wms.AddEvent(event);
                                }
                        }
                        return 0;
                case 'g':
                        if(buf[5]=='k'){
                                // goal kick
                                if(buf[10]=='l'){
                                        wms.SetPlayMode(PM_GOAL_KICK_LEFT);
                                        // ¼¼
                                        event.name = EV_GOAL_KICK_LEFT;
                                        wms.AddEvent(event);
                                }
                                else{
                                        wms.SetPlayMode(PM_GOAL_KICK_RIGHT);
                                        // ¼¼
                                        event.name = EV_GOAL_KICK_RIGHT;
                                        wms.AddEvent(event);
                                }
                        }
                        else{
                                if(buf[5]=='l'){
                                        // goal in left
                                        event.name = EV_GOAL_LEFT;
                                        sscanf(&buf[7],"%d",&temp);
                                        sprintf(tempbuf,"%d %d",temp,tm);
                                        event.value = tempbuf;
                                        wms.AddEvent(event);
                                }
                                else if(buf[5]=='r'){
                                        // goal in right
                                        event.name = EV_GOAL_RIGHT;
                                        sscanf(&buf[7],"%d",&temp);
                                        sprintf(tempbuf,"%d %d",temp,tm);
                                        event.value = tempbuf;
                                        wms.AddEvent(event);
                                }
                        }
                        return 0;
                case 'd':
                        // drop ball
                        wms.SetPlayMode(PM_DROP_BALL);
                        event.name = EV_DROP_BALL;
                        wms.AddEvent(event);
                        return 0;
                case 'o':
                        if(buf[8]=='l'){
                                wms.SetPlayMode(PM_OFFSIDE_LEFT);
                                event.name = EV_OFFSIDE_LEFT;
                                wms.AddEvent(event);
                        }
                        else{
                                wms.SetPlayMode(PM_OFFSIDE_RIGHT);
                                event.name = EV_OFFSIDE_RIGHT;
                                wms.AddEvent(event);
                        }
                        return 0;
                case 'c':
                        if(buf[12]=='l'){
                                wms.SetPlayMode(PM_CORNER_KICK_LEFT);
                                event.name = EV_CORNER_KICK_LEFT;
                                wms.AddEvent(event);
                        }
                        else{
                                wms.SetPlayMode(PM_CORNER_KICK_RIGHT);
                                event.name = EV_CORNER_KICK_RIGHT;
                                wms.AddEvent(event);
                        }
                        return 0;
                case 'f':
                        // ignore the foul message
                        if(buf[10]=='l'){
                                wms.SetPlayMode(PM_FREE_KICK_LEFT);
                                event.name = EV_FREE_KICK_LEFT;
                                wms.AddEvent(event);        
                        }
                        else{
                                wms.SetPlayMode(PM_FREE_KICK_RIGHT);
                                event.name = EV_FREE_KICK_RIGHT;
                                wms.AddEvent(event);
                        }
                        return 0;
                default:
                        return -1;
                }
        } // end if
        return -1;
}


int Parse::ParseSee(const char* Msg,WMSnapShot &wms)
{
        int tm,timeStamp; // time
        int br = 1; // brackets --> when meet left brackets, br++ ,else br-- 
        int token;
        int i,uniform;
        bool bOurSide;
        char buf[64];
        char num[16];   // store the uniform num
        float posx,posy,velx,vely,bodyang,neckang;
        EventT event;
        char *msg = (char *)Msg;
        
        // Get time
        if (GetToken(msg,buf) != WORD){
                // error,time should be here 
                return -1;
        }
        sscanf(buf,"%d",&tm);
        timeStamp = wms.GetTimeStamp();
        // ϢʱСڵǰʱ,old msg
        if( tm <= timeStamp){
            return -1;  // don't do anything
        }
        else if( tm > timeStamp+1 ){
            // some see msg packet are missed
            event.name = EV_MISS_SEE_MSG;
            // ʧӾϢeventֵ
            sprintf(buf,"%d",(tm - timeStamp + 1) );
            event.value = buf;
            wms.AddEvent(event);
        }
        
        wms.SetTimeStamp(tm);

        while(br){
                // Get next token              
                token = GetToken(msg,buf);
                if(token == LP) {
                        br++;
                        if(GetToken(msg,buf)!=LP){
                                return -1; // error, here should be LP
                        }
                        else{
                                GetToken(msg,buf);  // read ObjInfo
                                switch(buf[0]){
                                case 'b':  // ball
                                        if(GetToken(msg,buf)!=RP) 
                                            return -1; //error 
                                        sscanf(msg,"%f %f %f %f",&posx,&posy,&velx,&vely);
                                        
                                        // Set ball info
                                        wms.SetBallPositionX(posx);
                                        wms.SetBallPositionY(posy);
                                        wms.SetBallVelocityX(velx);
                                        wms.SetBallVelocityY(vely);
                                        wms.SetBallDataValidation(true);
                                        // skip over those tokens in sscanf
                                        for(i=0;i<5;i++) 
                                            token = GetToken(msg,buf); 

                                        if(token != RP) 
                                                return -1; // error
                                        else br--;

                                        break;
                                case 'p':  // player
                                        GetToken(msg,buf);  // Get team name

                                        bOurSide = (strcmp(buf,ourTeamName)==0)? true : false;

                                        if( !bOurSide && !bOppNameSet){
                                            // ״οֵĶ,
                                            event.name = EV_SEE_OPP_NAME;
                                            event.value = buf;      // Էvalue
                                            wms.AddEvent(event);
                                        }
                                        
                                        GetToken(msg,num);  // Get num
                                        sscanf(num,"%d",&uniform);
                                        
                                        GetToken(msg,buf); // Get next token, it should be RP or goalie
                                        if(buf[0] == 'g'){ 
                                                GetToken(msg,buf); // goalie
                                                // Ա,
                                               if ( bOurSide && !bOurGoalieSet){
                                                   event.name = EV_SEE_OUR_GOALIE;
                                                   event.value = num;
                                                   wms.AddEvent(event);
                                                   bOurGoalieSet = true;
                                               }
                                               else if( !bOurSide && !bOppGoalieSet){
                                                    event.name = EV_SEE_OPP_GOALIE;
                                                    event.value = num;
                                                    wms.AddEvent(event);
                                                    bOppGoalieSet = true;
                                               }
                                                
                                        }

                                        if(bOurSide){
                                       
                                                sscanf(msg,"%f %f %f %f %f %f",&posx,&posy,&velx,
                                                        &vely,&bodyang,&neckang);
                                                // Set teammate info
                                                wms.SetTeammatePositionX(posx,uniform);
                                                wms.SetTeammatePositionY(posy,uniform);
                                                wms.SetTeammateVelocityX(velx,uniform);
                                                wms.SetTeammateVelocityY(vely,uniform);
                                                wms.SetTeammateBodyAng(bodyang,uniform);
                                                wms.SetTeammateNeckAng(neckang,uniform);
                                                wms.SetTeammateDataValidation(true,uniform);
                                        }
                                        else{
                                      
                                                sscanf(msg,"%f %f %f %f %f %f",&posx,&posy,&velx,
                                                        &vely,&bodyang,&neckang);

                                                        
                                                // Set opponent info
                                                wms.SetOpponentPositionX(posx,uniform);
                                                wms.SetOpponentPositionY(posy,uniform);
                                                wms.SetOpponentVelocityX(velx,uniform);
                                                wms.SetOpponentVelocityY(vely,uniform);
                                                wms.SetOpponentBodyAng(bodyang,uniform);
                                                wms.SetOpponentNeckAng(neckang,uniform);
                                                wms.SetOpponentDataValidation(true,uniform);
                                        }  // read info of player

                                        for(i=0;i<7;i++)
                                           token = GetToken(msg,buf); // skip 7 words

                                        if(token!=RP)
                                           return -1; //error
                                        else br--;
                                        
                                        break;
                                case 'g':  // goal
                                        // leave it, do nothing
                                        for(i=0;i<5;i++)
                                            token = GetToken(msg,buf); // skip it
                                            
                                        if(token!=RP) 
                                            return -1; //error
                                            
                                        else br--;
                                        
                                        break;
                                default: // error
                                        return -1;
                                }
                        }

                }
                else if(token == RP){
                     br--;         
                }
                else {
cout<<"unexpected!"<<endl;                 
                    return -1;
                }
                    

        }  // end while

        return 0;
}

int Parse::ParseError(const char* Msg,WMSnapShot &wms)
{
    return 0;
}

int Parse::ParseWarning(const char* Msg,WMSnapShot &wms)
{
    return 0;
}

int Parse::ParseACK(const char* Msg,WMSnapShot &wms)
{
        char *pszMsg = (char *)Msg;
    
        switch( pszMsg[4] ){
        case 'c':
                // check_ball msg
                break;
        case 'e':
                if(pszMsg[5]=='y'){
                        // eye msg
                        if (pszMsg[9] =='n'){
                                 // (ok eye on)
                                //eye = true;
                        }
                        else{
                                // (ok eye off)
                                //eye = false;
                        }
                }
                else{
                        // ear msg
                        if (pszMsg[9] =='n'){
                                // (ok ear on)
                                //ear = true;
                        }
                        else{
                                // (ok ear off)
                                //ear = false;
                        }
                }
                break;
        case 'l':
                // look msg
                ParseSee(pszMsg+8,wms);
                break;
        case 's':
                // say msg
                break;
        default:
                break;
        }
        
        return 0;
}

int Parse::GetToken(char* &pszMsg,char *buf)
{
        // this function find tokens in msg and put them in buf,auto throw out \"
        // return value -- the name of tokens
        // delimiter --> ' ' and NULL
        // here assume the buf has enough length

        
        int length = 0;
        int rval ;

        while(*pszMsg == ' ') pszMsg++;  // skip blank

        switch(*pszMsg){
        case '(' :
                buf[length++] = *pszMsg++;
                rval = LP;
                break;
        case ')' :
                buf[length++] = *pszMsg++;
                rval = RP;
                break;
        case NULL :
                rval = DEL;
                break;
        default:
                if(*pszMsg == '\"') {
                        // string,skip \"
                        pszMsg++;
                        rval = STR;
                }
                else rval = WORD;

                while( (*pszMsg!=' ') && (*pszMsg != '(') && (*pszMsg != ')') 
                           && (*pszMsg!='\"') && (*pszMsg!= '\0') ){
                        buf[length++] = *pszMsg++;
                }  // copy token to buf

                if(*pszMsg == '\"') pszMsg++;
                break;
        }  // end switch

        buf[length] = '\0';
        
        return rval;
}

void     Parse::SetOurTeamName(const char *teamName)
{
        strcpy(ourTeamName,teamName);
}


int  Parse::ParseServerParam(const char* Msg,WMSnapShot &wms)
{
        // ¼¼
        EventT event;
        event.name = EV_GET_SERVER_PARAM;
        // parse the server parameters
        event.value = Msg;
        wms.AddEvent(event);
        return 0;
}

        
int  Parse::ParsePlayerParam(const char* Msg,WMSnapShot &wms)
{
        // ¼¼
        EventT event;
        event.name = EV_GET_PLAYER_PARAM;
        // parse the player parameters
        event.value = Msg;
        wms.AddEvent(event);
        return 0;
}

int  Parse::ParsePlayerType(const char* Msg,WMSnapShot &wms)
{
        // ¼¼
        EventT event;
        event.name = EV_GET_PLAYER_TYPE;
        // parse the player types
        event.value = Msg;
        wms.AddEvent(event);
        return 0;
}

