/***************************************************************************
                                  TeamModel.h 
                             -------------------
    begin                : 2001
    copyright            : (C)2001-2003 Multi-Agent Systems Laboratory,
                           University of Science and Technology of China 
                           Zhanxiang Huang, Yang Yang and Xiaoping Chen.
    email                : huangzx@mail.ustc.edu.cn
 ***************************************************************************/

#ifndef     _TEAMMODEL_H_
#define     _TEAMMODEL_H_


#include <vector>
#include <set>
#include "Types.h"
#include "WorldModel.h"
#include "ServerSetting.h"
#include "Logger.h"

using namespace std;

class Player{
public:
        Player();
        
        // interface for setting the player parameters
        bool SetMaxSpeed(const double s);
        bool SetMaxAcceleration(const double a);
        bool SetMaxKick(const double k);
        bool SetOppDensity(const double d);
        
        bool SetMaxDribbleTime(const int dt);
        bool SetShootCount(const int sc);
        bool SetPassCount(const int pc);
        bool SetDribbleCount(const int dc);
        bool SetGoalCount(const int gc);
        
        // interface for Getting the player parameters
        double GetMaxSpeed();
        double GetMaxAcceleration();
        double GetMaxKick();
        double GetOppDensity();
        
        int GetMaxDribbleTime();
        int GetShootCount();
        int GetPassCount();
        int GetDribbleCount();
        int GetGoalCount();
        
        int SetRole(PlayerRoleT r);
        PlayerRoleT GetRole();
        PlayerActS GetAction();
        void OutPlayerRole();


protected:
        double      maxSpeed;    //the Maximum speed the player ever achieved
        double      maxAcceleration; //the Maximum acceleration the player ever achieved
        double      maxKick; //the Maximum ball speed leaving the player
        double      oppDensity;
        
        int         maxDribbleTime; //the Maximum time the player drible with interference
        int         shootCount; //the number the player shoot
        int         passCount;  //the number the player pass ball
        int         dribbleCount; //the number the player dribble
        int         goalCount;  //the number the player goal
        
        ActionT     action; // the parameter describe the posible action the player now taking
        int         passObjectCount[MAX_PLAYER_NUMBER];
        int         passObjectSuccessCount[MAX_PLAYER_NUMBER];
        int         passObjectRate[MAX_PLAYER_NUMBER];
        int         passObjectSuccessRate[MAX_PLAYER_NUMBER];

        PlayerRoleT role;
        
public:
        vector<BallPossessionStruct>    possessionRecord;
};



class Formation{
public:
        Formation();
        
        VecPos      expectedFormation[MAX_PLAYER_NUMBER];
        
        int         goalie;

        VecPos      rightestPlayerPos;
        VecPos      leftestPlayerPos;
        VecPos      topestPlayerPos;
        VecPos      bottomestPlayerPos;

        set<int>    forwardLeft;
        set<int>    forwardCenter;
        set<int>    forwardRight;
        set<int>    midfieldLeft;
        set<int>    midfieldCenter;
        set<int>    midfieldRight;
        set<int>    backLeft;
        set<int>    backCenter;
        set<int>    backRight;
        
        int         forwardNumLast;
        int         middleNumLast;
        int         backNumLast;
        int         forwardNum;
        int         middleNum;
        int         backNum;
};



class Team{
public:
        Team();
        int         SetWorldModelAndParameters(WorldModel* cwm, ServerSetting* css, PlayerSetting* cps, PlayerTypeSetting* cpts);
        int         UpdatePlayerExpectedPosition(SideT side, int ct, int interval);
        void        OutExpectedFormation();
        SideT       GetSide(TeamT t);
        void        OutBallPossessionRecord();
        void        UpdatePlayerPositionModel(SideT side, int ct, int interval);
        void        SetLogger(Logger *log);
        int         GetControlBallTime();
        double      GetControlBallRatio();


        Player p[MAX_PLAYER_NUMBER];
        Formation f;
        
        //positionModel[X|Y][f|m|b][UNUM][PREFIX]
        double  positionModel[2][3][MAX_PLAYER_NUMBER][POSITION_MODEL_PREFIX_NUMBER];

        vector<BallPossessionStruct>    possessionRecord;
        
        int role2PlayerMap[MAX_PLAYER_NUMBER];
        

private:
        WorldModel* wm;
        ServerSetting* ss;
        PlayerSetting* ps;
        PlayerTypeSetting* pts;
        Logger  *l;
        

        //P[X|Y][f|m|b][player][][]
        double  pp[2][3][MAX_PLAYER_NUMBER][POSITION_MODEL_PREFIX_NUMBER][POSITION_MODEL_PREFIX_NUMBER];
        double  ll[2][3][MAX_PLAYER_NUMBER][POSITION_MODEL_PREFIX_NUMBER];
        double  x[POSITION_MODEL_PREFIX_NUMBER];
        double  xx[POSITION_MODEL_PREFIX_NUMBER];
        double  temp[2][3][POSITION_MODEL_PREFIX_NUMBER];
        double  temp2[2][3];
        double  temp3[POSITION_MODEL_PREFIX_NUMBER][POSITION_MODEL_PREFIX_NUMBER];
        double  temp4[POSITION_MODEL_PREFIX_NUMBER][POSITION_MODEL_PREFIX_NUMBER];
        double  e[2][3];
        double  temp5[2][3];
};


#endif



