/***************************************************************************
                                  WECoach.h 
                             -------------------
    begin                : 2001
    copyright            : (C)2001-2003 Multi-Agent Systems Laboratory,
                           University of Science and Technology of China 
                           Zhanxiang Huang, Yang Yang and Xiaoping Chen.
    email                : huangzx@mail.ustc.edu.cn
 ***************************************************************************/

#ifndef _WECOACH_H_
#define _WECOACH_H_

#include <list>
#include <stdio.h>
#include <string>
#include "ServerSetting.h"
#include "WorldModel.h"
#include "Netif.h"
#include "Types.h"
#include "Parse.h"
#include "BasicCoach.h"
#include "Logger.h"
#include "TeamModel.h"

using namespace std;

class WECoach: public BasicCoach{
public:
        // construct and initiation
        WECoach();
        WECoach(UDPSocket *s , Parse *parse, string &ourTeamName, Logger* cl, SideT side);
        int InitFormPara(); // ʼβ

        // update functions
        int UpdateRoutine();
        int UpdateTeamModel();
        int UpdateFormationStat();
        int UpdateMaxSpeed();
        int UpdateHoldStat();
        int UpdatePassStat();
        int UpdateDribbleStat();
        int UpdateShootStat();
        
        
        // send command functions
        int SendCommand();
        string ChangePlayerTypeCommand(int unum, int playerType);
        int ChangeWEPlayerType();
        string HomeCommand(FormationT fm,int dhs,int dvs,int dl,
                        int chs,int cvs,int cl,int fhs,int fvs,int fl);
        string PlayMode2String(PlayModeT pm);
        string BallPos2String(Region ballPos);
        string PlayerPos2String(TeamT team, vector<int> unum, int min, int max, Region playerPos);
        string PlayerAction2String(PlayerActS playerAct);
        string PlayerAction2String(ActionMode mode, TeamT team, vector<int> unum, PlayerActS playerAct);
        int    SentDefineMsg();
        int    Int2RegionChar(int x);
        string GetRuleName();



        // decision functions
        int Decision(bool online);

        int DecidePlayerType();
        int DecidePositioning();
        int DoAttackDecision();
        void DecideTeamplay();
        vector<string> GenOurTactic(FormationT fm,Tactic tac,int LifeSpan);
        int  DecideTeachTeamplay();

        int DoInfoDecision();
        
        int DecideMark();
        int DoDefendDecision();


        // analyze functions
        int Analyze();
        BallPossessionT GetBallPossessionState(int tm);
        int UpdateBallPossessionState();
        int AnalyzePlayerRole(TeamT t, int uniformNumber);
        int AnalyzePlayerRoles();
        int AnalyzeBallPossession(BallPossessionT bps);
        int AnalyzePlayerAction(BallPossessionT bps);
        int AnalyzeTeammateAction();
        int AnalyzeOpponentAction();
        bool NeedToDecideNewPosition();
        FormationT GetFormation(TeamT tm);
        

        // supervise functions
        int Supervise();
        

        // memory and judge functions
        BallPossessionT IsBallPossessionStateChange();

private:
        Logger* l;
        Team    mm;    // our team model, mm Is short for 'mate team model'
        Team    om;    // opponent team model, om Is short for 'opp team model'
        
        int     iSayInterval;    // count the Minimum interval between send command actions
        
        list<Command>    sendMsgQ;   // queue of message sending from coach to server
        
        // coach memory
        int     lastDecisionTime;
        BallPossessionT    lastBallPossessionState;
        BallPossessionT    currentBallPossessionState;
        bool    bGoalieSet;
        bool    bPlayerRolesSet;
        bool    bDecidePlayerType;
        bool    bDecidePositioning;
        int     iOnlineChangePlayerTypeCount;
        
        // formation parameter use for formation generation
        // there 3 styles :
        // 0. protrude
        // 1. plain
        // 2. recess
        int H2[3][2];  // 2Ա 
        int V2[3][2];  // 2Ա ֲ
        int H3[3][3];  // 3Ա 
        int V3[3][3];  // 3Ա ֲ
        int H4[3][4];  // 4Ա 
        int V4[3][4];  // 4Ա ֲ
        int H5[3][5];  // 5Ա 
        int V5[3][5];  // 5Ա ֲ
        int forwardLine;
        int middleLine;
        int backLine;
        
        int checkPossessionRecordIndex;
        
        int     rowNum, volNum;//read from file
        double  rowWidth, volWidth;
        
public:
        bool    receiveMsgMark;
        int     ruleName;
        int     proccessRuleMark;
        vector<RuleRecord>  ruleRecordQ;
        int     superviseRuleMark;

};

#endif

