/***************************************************************************
                                  WorldModel.C 
                             -------------------
    begin                : 2001
    copyright            : (C)2001-2003 Multi-Agent Systems Laboratory,
                           University of Science and Technology of China 
                           Zhanxiang Huang, Yang Yang and Xiaoping Chen.
    email                : huangzx@mail.ustc.edu.cn
 ***************************************************************************/

#include "WorldModel.h"

 
//****************************************************
//class Object
//****************************************************
Object::Object(){
       SetDataValidation(false);
}

// interface of Setting informations

bool  Object::SetPosition(const VecPos &pos)
{
   position = pos;
   return true;
}


bool  Object::SetPositionX( const double x)
{
      position.SetX(x);
      return true;
}


bool   Object::SetPositionY(const double y)
{
        position.SetY(y);
        return true;
}


bool  Object::SetVelocity(const VecVel &vel)
{
        velocity = vel;
        return true;
}

bool  Object::SetVelocityX(const double x)
{
        velocity.SetX(x);
        return true;
}


bool  Object::SetVelocityY(const double y)
{
        velocity.SetY(y);
        return true;
}

//  interface of Getting informations
        
VecPos   Object::GetPosition() const
{
        return position;        
}


double   Object::GetPositionX() const
{
        return position.GetX();
}


double   Object::GetPositionY() const
{
        return position.GetY();
}


 VecVel   Object::GetVelocity() const
 {
        return velocity;
}

double    Object:: GetVelocityX()  const
{
        return velocity.GetX();
}


double  Object::GetVelocityY() const
{
        return velocity.GetY();
}

     
double  Object::GetSpeed() const
{
        return velocity.GetMagnitude();
}


 double  Object::GetDistanceTo(const Object o)
 {
       // Getting distance to object o 
       VecPos pos = o.GetPosition();
       return position.GetDistanceTo( pos );

}

bool   Object::IsDataValid()  const
{
        return bDataValid;
}


void  Object::SetDataValidation(bool b)
{
        // Чλ
        bDataValid = b;
}


//****************************************************
//class BallObject
//****************************************************
BallObject::BallObject()
{
        
}
       

//****************************************************
//class PlayerObject
//****************************************************
PlayerObject::PlayerObject()
{
}

        
bool  PlayerObject::SetBodyAng(const AngDeg d)
{
        bodyAng = d;
        return true;
}

bool  PlayerObject::SetNeckAng(const AngDeg d)
{
        neckAng = d;
        return true;
}

        
AngDeg PlayerObject::GetBodyAng() const
{
        return bodyAng;
}


AngDeg PlayerObject::PlayerObject::GetNeckAng() const
{
        return neckAng;
}


//****************************************************
//Class WMSnapShot
//****************************************************
WMSnapShot::WMSnapShot()
{
    SetAllDataInvalidate();
}


// interface of Setting informations
bool    WMSnapShot::SetPlayMode(const PlayModeT mode)
{
        playmode = mode;
        return true;
}


bool    WMSnapShot::SetOurScore(const int score)
{
        ourScore = score;
        return true;
}


bool    WMSnapShot::SetOppScore(const int score)
{
        oppScore = score;
        return true;
}

bool    WMSnapShot::SetBallPosition(const VecPos &pos)
{
        return ball.SetPosition(pos) ;
}


bool    WMSnapShot::SetBallPositionX(const double x) 
{
        return ball.SetPositionX(x);
}

bool    WMSnapShot::SetBallPositionY(const double y) 
{
       return ball.SetPositionY(y); 
}


bool    WMSnapShot::SetBallVelocity(const VecVel &vel)
{
        return ball.SetVelocity(vel);
}


bool    WMSnapShot::SetBallVelocityX(const double x)
{
        return ball.SetVelocityX(x);
}

bool    WMSnapShot::SetBallVelocityY(const double y)
{
      return ball.SetVelocityY(y);
};


// Set informations for teammate i
bool    WMSnapShot::SetTeammatePosition( const VecPos &pos,  const int uniform)
{
        return teammates[uniform-1].SetPosition(pos);
}


bool    WMSnapShot::SetTeammatePositionX( const double x,  const int uniform)
{                   
        return teammates[uniform-1].SetPositionX(x);
}


bool    WMSnapShot::SetTeammatePositionY(const double y, const int uniform)
{
         return teammates[uniform-1].SetPositionY(y);
}


bool    WMSnapShot::SetTeammateVelocity(const VecVel &vel,  const int uniform)
{
        return teammates[uniform-1].SetVelocity(vel);
}


bool    WMSnapShot::SetTeammateVelocityX(const double x,  const int uniform)
{
        return teammates[uniform-1].SetVelocityX(x);
}


bool    WMSnapShot::SetTeammateVelocityY(const double y,  const int uniform)
{
        return teammates[uniform-1].SetVelocityY(y);
}


bool    WMSnapShot::SetTeammateBodyAng(const AngDeg ang, const  int uniform)
{
        return teammates[uniform-1].SetBodyAng(ang);
}


bool    WMSnapShot::SetTeammateNeckAng(const AngDeg ang, const  int uniform)
{
        return teammates[uniform-1].SetNeckAng(ang);
}


bool    WMSnapShot::SetOpponentPosition(VecPos &pos,  const int uniform)
{
        return opponents[uniform-1].SetPosition(pos);
}


bool    WMSnapShot::SetOpponentPositionX( const double x,  const int uniform)
{
        return opponents[uniform-1].SetPositionX(x);
}


bool    WMSnapShot::SetOpponentPositionY(const double y, const int uniform)
{
        return opponents[uniform-1].SetPositionY(y);
}


bool    WMSnapShot::SetOpponentVelocity(const VecVel &vel,  const int uniform)
{
        return opponents[uniform-1].SetVelocity(vel);
}


bool    WMSnapShot::SetOpponentVelocityX(const double x,  const int uniform)
{
        return opponents[uniform-1].SetVelocityX(x);
}


bool    WMSnapShot::SetOpponentVelocityY(const double y,  const int uniform)
{
        return opponents[uniform-1].SetVelocityY(y);
}


bool    WMSnapShot::SetOpponentBodyAng(const AngDeg ang,  const int uniform)
{
        return opponents[uniform-1].SetBodyAng(ang);
}


bool    WMSnapShot::SetOpponentNeckAng(const AngDeg ang, const  int uniform)
{
        return opponents[uniform-1].SetNeckAng(ang);
}

bool    WMSnapShot::SetTimeStamp(const int t)
{
        timeStamp = t;
        return true;
        
}
    
bool    WMSnapShot::AddEvent(const EventT event)
{
        events.push_back(event);
        return true;
}
    
                  
//interface of Getting informations
PlayModeT       WMSnapShot::GetPlayMode() const
{
        return playmode;
}


int   WMSnapShot::GetOurScore() const
{
        return ourScore;
}


int   WMSnapShot::GetOppScore() const
{
        return oppScore;
}

        
VecPos WMSnapShot::GetBallPosition() const
{
        return ball.GetPosition();
}

double  WMSnapShot::GetBallPositionX() const
{
        return ball.GetPositionX();
}

double   WMSnapShot::GetBallPositionY() const
{
        return ball.GetPositionY();
}
        
        
VecVel   WMSnapShot::GetBallVelocity() const
{
        return ball.GetVelocity();
}

double   WMSnapShot::GetBallVelocityX() const
{
        return ball.GetVelocityX();
}


double   WMSnapShot::GetBallVelocityY() const
{
        return ball.GetVelocityY();
}

double  WMSnapShot::GetBallSpeed() const
{
        return ball.GetSpeed();
}


        
VecPos  WMSnapShot::GetTeammatePosition( const int uniform) const
{
        return teammates[uniform-1].GetPosition();
}


double   WMSnapShot::GetTeammatePositionX( const int uniform) const
{
        return teammates[uniform-1].GetPositionX();
}


double   WMSnapShot::GetTeammatePositionY( const int uniform) const
{
        return teammates[uniform-1].GetPositionY();
}


VecVel  WMSnapShot::GetTeammateVelocity( const  int uniform) const
{
        return teammates[uniform-1].GetVelocity();
}


double   WMSnapShot::GetTeammateVelocityX( const int uniform) const
{
        return teammates[uniform-1].GetVelocityX();
}


double   WMSnapShot::GetTeammateVelocityY( const int uniform) const
{
        return teammates[uniform-1].GetVelocityY();
}


double   WMSnapShot::GetTeammateSpeed( const int uniform) const
{
        return teammates[uniform-1].GetSpeed();
}


AngDeg    WMSnapShot::GetTeammateBodyAng(const int uniform) const
{
        return teammates[uniform-1].GetBodyAng();
}


AngDeg    WMSnapShot::GetTeammateNeckAng( const int uniform) const
{
        return teammates[uniform-1].GetNeckAng();
}

        
VecPos  WMSnapShot::GetOpponentPosition(  const  int uniform) const
{
        return opponents[uniform-1].GetPosition();
}


double WMSnapShot::GetOpponentPositionX(  const  int uniform)  const
{
        return opponents[uniform-1].GetPositionX();
}


double WMSnapShot::GetOpponentPositionY(  const int uniform)  const
{
        return opponents[uniform-1].GetPositionY();
}


VecVel WMSnapShot::GetOpponentVelocity( const  int uniform)  const
{
        return opponents[uniform-1].GetVelocity();
}


double WMSnapShot::GetOpponentVelocityX( const int uniform)  const
{
        return opponents[uniform-1].GetVelocityX();
}


double WMSnapShot::GetOpponentVelocityY(  const int uniform)  const
{
        return opponents[uniform-1].GetVelocityY();
}


double WMSnapShot::GetOpponentSpeed( const int uniform)  const
{
        return opponents[uniform-1].GetSpeed();
}


AngDeg  WMSnapShot::GetOpponentBodyAng(  const int uniform )  const
{
        return opponents[uniform-1].GetBodyAng();
}


AngDeg  WMSnapShot::GetOpponentNeckAng( const int uniform)  const
{
        return opponents[uniform-1].GetNeckAng();
}

int     WMSnapShot::GetTimeStamp() const
{
        return timeStamp;
}
    
list<EventT>& WMSnapShot::GetEvents()
{
    return events;    
}


// Interface that Setting data validation
void WMSnapShot::SetBallDataValidation(bool b)
{
    ball.SetDataValidation(b);
}


void WMSnapShot::SetTeammateDataValidation(bool b, const int uniform) 
{
    teammates[uniform-1].SetDataValidation(b);
}


void WMSnapShot::SetOpponentDataValidation(bool b, const int uniform)
{
    opponents[uniform-1].SetDataValidation(b);
}

// WMSnopShotеϢΪfalse, play_mode-->PM_ILLEGAL
// ͬʱeventsÿ, ourScore,oppScore = 0; timeStamp = 0;

void WMSnapShot::SetAllDataInvalidate()
{
    ball.SetDataValidation(false);
    
    for(int i=0;i<MAX_PLAYER_NUMBER;i++){
        teammates[i].SetDataValidation(false);
        opponents[i].SetDataValidation(false);        
    }

    playmode = PM_ILLEGAL;
    
    ourScore = 0;
    oppScore = 0;
    
    timeStamp = 0;
    
    events.clear();
    
}

        
// Interface that Getting data validation
bool  WMSnapShot::GetBallDataValidation() const
{
    return ball.IsDataValid();
}


bool  WMSnapShot::GetTeammateDataValidation(const int uniform) const
{
    return teammates[uniform].IsDataValid();
}


bool  WMSnapShot::GetOpponentDataValidation(const int uniform) const
{
    return opponents[uniform].IsDataValid();
}

void  WMSnapShot::clearEvent(){
    events.clear();   
}




//****************************************************
//Class WorldModel
//****************************************************

WorldModel::WorldModel()
{
    ctm = 0;
    
    WMSnapShot tmpWMS;
    
    field.empty();
    
    // insert a world snapshot at time 0

    tmpWMS.SetAllDataInvalidate();
    tmpWMS.SetOurScore(0);
    tmpWMS.SetOppScore(0);
    tmpWMS.SetPlayMode(PM_BEFORE_KICK_OFF);
    field.push_back(tmpWMS);

    bOurGoalieUniformSet = false;
    bOppGoalieUniformSet = false;
    
    bOppTeamNameSet = false;
    bOurTeamNameSet = false;
}

void WorldModel::SetLogger(Logger *log)
{
    l = log;
}

bool    WorldModel::SetCurrentTime(const int tm)
{
    ctm = tm;
    return true;
}


// interface of Setting informations
bool    WorldModel::SetPlayMode(const PlayModeT mode, int tm)
{
        return field[tm].SetPlayMode(mode);
}


bool    WorldModel::SetOurScore(const int score, int tm)
{
       return field[tm].SetOurScore(score);
}


bool    WorldModel::SetOppScore(const int score, int tm)
{
        return field[tm].SetOppScore(score);
}


bool    WorldModel::SetBallPosition(const VecPos &pos, int tm)
{
        return field[tm].SetBallPosition(pos);
}


bool    WorldModel::SetBallPositionX(const double x, int tm) 
{
        return field[tm].SetBallPositionX(x);
}

bool    WorldModel::SetBallPositionY(const double y, int tm) 
{
       return field[tm].SetBallPositionY(y);
}


bool    WorldModel::SetBallVelocity(const VecVel &vel, int tm)
{
        return field[tm].SetBallVelocity(vel);
}


bool    WorldModel::SetBallVelocityX(const double x, int tm)
{
        return field[tm].SetBallVelocityX(x);
}

bool    WorldModel::SetBallVelocityY(const double y, int tm)
{
      return field[tm].SetBallVelocityY(y);
}


// Set informations for teammate i
bool    WorldModel::SetTeammatePosition( const VecPos &pos,  const int uniform, int tm)
{
        return field[tm].SetTeammatePosition(pos,uniform);
}


bool    WorldModel::SetTeammatePositionX( const double x,  const int uniform, int tm)
{
        return field[tm].SetTeammatePositionX(x,uniform);
}


bool    WorldModel::SetTeammatePositionY(const double y, const int uniform, int tm)
{
         return field[tm].SetTeammatePositionY(y,uniform);
}


bool    WorldModel::SetTeammateVelocity(const VecVel &vel,  const int uniform, int tm)
{
        return field[tm].SetTeammateVelocity(vel,uniform);
}


bool    WorldModel::SetTeammateVelocityX(const double x,  const int uniform, int tm)
{
        return field[tm].SetTeammateVelocityX(x,uniform);
}


bool    WorldModel::SetTeammateVelocityY(const double y,  const int uniform, int tm)
{
        return field[tm].SetTeammateVelocityY(y,uniform);
}


bool    WorldModel::SetTeammateBodyAng(const AngDeg ang, const  int uniform, int tm)
{
        return field[tm].SetTeammateBodyAng(ang,uniform);
}


bool    WorldModel::SetTeammateNeckAng(const AngDeg ang, const  int uniform, int tm)
{
        return field[tm].SetTeammateNeckAng(ang,uniform);
}


bool    WorldModel::SetOpponentPosition(VecPos &pos,  const int uniform, int tm)
{
        return field[tm].SetOpponentPosition(pos,uniform);
}


bool    WorldModel::SetOpponentPositionX( const double x,  const int uniform, int tm)
{
        return field[tm].SetOpponentPositionX(x, uniform);
}


bool    WorldModel::SetOpponentPositionY(const double y, const int uniform, int tm)
{
        return field[tm].SetOpponentPositionY(y,uniform);
}


bool    WorldModel::SetOpponentVelocity(const VecVel &vel,  const int uniform, int tm)
{
        return field[tm].SetOpponentVelocity(vel,uniform);
}


bool    WorldModel::SetOpponentVelocityX(const double x,  const int uniform, int tm)
{
        return field[tm].SetOpponentVelocityX(x,uniform);
}


bool    WorldModel::SetOpponentVelocityY(const double y,  const int uniform, int tm)
{
        return field[tm].SetOpponentVelocityY(y,uniform);
}


bool    WorldModel::SetOpponentBodyAng(const AngDeg ang,  const int uniform, int tm)
{
        return field[tm].SetOpponentBodyAng(ang,uniform);
}


bool    WorldModel::SetOpponentNeckAng(const AngDeg ang, const  int uniform, int tm)
{
        return field[tm].SetOpponentNeckAng(ang,uniform);
}

bool    WorldModel::AddEvent(const EventT event, int tm)
{
        return field[tm].AddEvent(event);
}

              
//interface of Getting informations
PlayModeT       WorldModel::GetPlayMode(int tm) const
{
        return field[tm].GetPlayMode();
}


int   WorldModel::GetOurScore(int tm) const
{
        return field[tm].GetOurScore();
}


int   WorldModel::GetOppScore(int tm) const
{
        return field[tm].GetOppScore();
}

        
VecPos WorldModel::GetBallPosition(int tm) const
{
        return field[tm].GetBallPosition();
}

double  WorldModel::GetBallPositionX( int tm) const
{
        return field[tm].GetBallPositionX();
}

double   WorldModel::GetBallPositionY( int tm) const
{
        return field[tm].GetBallPositionY();
}
        
        
VecVel   WorldModel::GetBallVelocity(int tm) const
{
        return field[tm].GetBallVelocity();
}


double   WorldModel::GetBallVelocityX( int tm) const
{
        return field[tm].GetBallVelocityX();
}


double   WorldModel::GetBallVelocityY( int tm) const
{
        return field[tm].GetBallVelocityY();
}

double  WorldModel::GetBallSpeed( int tm) const
{
        return field[tm].GetBallSpeed();
}
        
VecPos  WorldModel::GetTeammatePosition( const int uniform, int tm) const
{
        return field[tm].GetTeammatePosition(uniform);
}


double   WorldModel::GetTeammatePositionX( const int uniform, int tm) const
{
        return field[tm].GetTeammatePositionX(uniform);
}


double   WorldModel::GetTeammatePositionY( const int uniform, int tm) const
{
        return field[tm].GetTeammatePositionY(uniform);
}


VecVel  WorldModel::GetTeammateVelocity( const  int uniform, int tm) const
{
        return field[tm].GetTeammateVelocity(uniform);
}


double   WorldModel::GetTeammateVelocityX( const int uniform, int tm) const
{
        return field[tm].GetTeammateVelocityX(uniform);
}


double   WorldModel::GetTeammateVelocityY( const int uniform, int tm) const
{
        return field[tm].GetTeammateVelocityY(uniform);
}


double   WorldModel::GetTeammateSpeed( const int uniform, int tm) const
{
        return field[tm].GetTeammateSpeed(uniform);
}


AngDeg    WorldModel::GetTeammateBodyAng(const int uniform, int tm) const
{
        return field[tm].GetTeammateBodyAng(uniform);
}


AngDeg    WorldModel::GetTeammateNeckAng( const int uniform, int tm) const
{
        return field[tm].GetTeammateNeckAng(uniform);
}

        
VecPos    WorldModel::GetOpponentPosition(  const  int uniform, int tm) const
{
        return field[tm].GetOpponentPosition(uniform);
}


double      WorldModel::GetOpponentPositionX(  const  int uniform, int tm)  const
{
        return field[tm].GetOpponentPositionX(uniform);
}


double      WorldModel::GetOpponentPositionY(  const int uniform, int tm)  const
{
        return field[tm].GetOpponentPositionY(uniform);
}


VecVel     WorldModel::GetOpponentVelocity( const  int uniform, int tm)  const
{
        return field[tm].GetOpponentVelocity(uniform);
}


double      WorldModel::GetOpponentVelocityX( const int uniform, int tm)  const
{
        return field[tm].GetOpponentVelocityX(uniform);
}


double      WorldModel::GetOpponentVelocityY(  const int uniform, int tm)  const
{
        return field[tm].GetOpponentVelocityY(uniform);
}


double      WorldModel::GetOpponentSpeed( const int uniform, int tm)  const
{
        return field[tm].GetOpponentSpeed(uniform);
}


AngDeg    WorldModel::GetOpponentBodyAng(  const int uniform, int tm )  const
{
        return field[tm].GetOpponentBodyAng(uniform);
}


AngDeg    WorldModel::GetOpponentNeckAng( const int uniform, int tm)  const
{
        return field[tm].GetOpponentNeckAng(uniform);
}


bool   WorldModel:: UpdateWorldModel( WMSnapShot  &wms) 
{
        ctm = wms.GetTimeStamp();
        
        field.push_back(wms);        
        
        return true;
}

int  WorldModel::GetCurrentTime() const
{
    return ctm;
}

list<EventT>& WorldModel::GetEvents(int tm)
{
    return field[tm].GetEvents();
}


WMSnapShot WorldModel::GetWMSnapShot(const int tm) const
{
    return field[tm];
}



bool   WorldModel::IsOurGoalieNumSet()  const
{
        return bOurGoalieUniformSet;
}


bool    WorldModel::IsOppGoalieNumSet()  const
{
        return bOppGoalieUniformSet;
}

// ԱºţͬʱbGoalieUniformSetΪ true

void    WorldModel::SetOurGoalieNum(const int unum)
{
        ourGoalieUniform = unum;
        bOurGoalieUniformSet = true;
}

void    WorldModel::SetOppGoalieNum(const int num)
{
        oppGoalieUniform = num;
        bOppGoalieUniformSet = true;
}

int     WorldModel::GetOurGoalieNum()
{
    return ourGoalieUniform;
}

int     WorldModel::GetOppGoalieNum()
{
    return oppGoalieUniform;
}
        
bool    WorldModel::IsOurTeamNameSet() const
{
        return bOurTeamNameSet;
}


bool    WorldModel::IsOppTeamNameSet() const 
{
        return bOppTeamNameSet;
}


// ͬʱbTeamNameSetΪtrue
void    WorldModel::SetOurTeamName(string &teamName)
{
        ourTeamName = teamName;
        bOurTeamNameSet = true;
}


void    WorldModel::SetOppTeamName(string &teamName)
{
        oppTeamName = teamName;
        bOppTeamNameSet = true;
}

void    WorldModel::SetOurSide(SideT side)
{
        ourSide = side;
        if( side == FIELD_SIDE_L)
            SetOppSide(FIELD_SIDE_R);
        else
            SetOppSide(FIELD_SIDE_L);
}

void    WorldModel::SetOppSide(SideT side)
{
        oppSide = side;
}
        
SideT   WorldModel::GetOurSide()
{
        return ourSide;
}

SideT   WorldModel::GetOppSide()
{
        return oppSide;
}


void    WorldModel::SetServerParameters(ServerSetting* css, PlayerSetting* cps, PlayerTypeSetting* cpts)
{
        ss = css;
        ps = cps;
        pts = cpts;
}


double  WorldModel::GetPlayerPositionX(SideT side, const  int uniform, int tm)  const
{
        if( side == ourSide )
            return GetTeammatePositionX(uniform, tm);
        else
            return GetOpponentPositionX(uniform, tm);
}


double  WorldModel::GetPlayerPositionY(SideT side, const int uniform, int tm)  const
{
        if( side == ourSide )
            return GetTeammatePositionY(uniform, tm);
        else
            return GetOpponentPositionY(uniform, tm);
}


VecVel  WorldModel::GetPlayerVelocity(SideT side, const  int uniform, int tm)  const
{
        if( side == ourSide )
            return GetTeammateVelocity(uniform, tm);
        else
            return GetOpponentVelocity(uniform, tm);
}


double  WorldModel::GetOpponentVelocityX(SideT side, const int uniform, int tm)  const
{
        if( side == ourSide )
            return GetTeammateVelocityX(uniform, tm);
        else
            return GetOpponentVelocityX(uniform, tm);
}


double  WorldModel::GetOpponentVelocityY(SideT side,  const int uniform, int tm)  const
{
        if( side == ourSide )
            return GetTeammateVelocityY(uniform, tm);
        else
            return GetOpponentVelocityY(uniform, tm);
}


double  WorldModel::GetOpponentSpeed(SideT side, const int uniform, int tm)  const
{
        if( side == ourSide )
            return GetTeammateSpeed(uniform, tm);
        else
            return GetOpponentSpeed(uniform, tm);
}


AngDeg  WorldModel::GetOpponentBodyAng(SideT side,  const int uniform, int tm )  const
{
        if( side == ourSide )
            return GetTeammateBodyAng(uniform, tm);
        else
            return GetOpponentBodyAng(uniform, tm);
}


AngDeg  WorldModel::GetOpponentNeckAng(SideT side, const int uniform, int tm)  const
{
        if( side == ourSide )
            return GetTeammateNeckAng(uniform, tm);
        else
            return GetOpponentNeckAng(uniform, tm);
}



vector<int>     WorldModel::GetOurBallOwner(int tm)
{
    vector<int> ballOwner;
    for(int i=0;i<MAX_PLAYER_NUMBER;i++)
    {
        if(GetTeammatePosition(i+1, tm).GetDistanceTo(GetBallPosition(tm))
            <= (pts->type[0].kickable_margin+pts->type[0].player_size) )
        {
            ballOwner.push_back(i+1);
        }
    }
    return ballOwner;
}

vector<int>     WorldModel::GetOppBallOwner(int tm)
{
    vector<int> ballOwner;
    for(int i=0;i<MAX_PLAYER_NUMBER;i++)
    {
        if(GetOpponentPosition(i+1, tm).GetDistanceTo(GetBallPosition(tm))
            < (pts->type[0].kickable_margin+pts->type[0].player_size) )
            ballOwner.push_back(i+1);
    }
    return ballOwner;
}


void        WorldModel::OutBallOwner(vector<int> ballOwner)
{
    vector<int>::iterator iter;
    cout<<"at time "<<GetCurrentTime()<<" ball owners are : ";
    for( iter=ballOwner.begin(); iter!=ballOwner.end(); iter++)
    {
        cout<<(*iter)<<" ";
    }
    cout<<endl;
}


bool WorldModel::IsBallKickableForTeammate(int uniform, int tm)
{
    double d = GetTeammatePosition(uniform, tm).GetDistanceTo(GetBallPosition(tm));
    if( d < (pts->type[0].kickable_margin+pts->type[0].player_size) ) return true;
    return false;
}

bool WorldModel::IsBallKickableForOpponent(int uniform, int tm)
{
    double d = GetOpponentPosition(uniform, tm).GetDistanceTo(GetBallPosition(tm));
    if( d < (pts->type[0].kickable_margin+pts->type[0].player_size) ) return true;
    return false;
}

bool WorldModel::IsBallKickableForPlayer(SideT side, int uniform, int tm)
{
    if( side == ourSide )
        return IsBallKickableForTeammate(uniform, tm);
    else
        return IsBallKickableForOpponent(uniform, tm);
}


// if a player's position has not changed in the recent holdball_threshold cycle
// then we will deem that the player Is holding ball.

double WorldModel::IsPass(VecPos ballPos,VecVel ballVel,int tm,PlayerActS *act)
{
    VecPos _ballPos = ballPos;
    VecPos _ballVel(ballVel.GetX(),ballVel.GetY());
    bool notFind = true;
    float cyclePass = 0;
    
    while(notFind && cyclePass<PASS_CYCLE_MAX_THRESHOLD)
    {
        //calculate the new ball position
        _ballPos += _ballVel;
        _ballVel.SetMagnitude(_ballVel.GetMagnitude()*ss->ball_decay);
        cyclePass++;
        //calculate the mate who can reach the ball
        for(int i=0;i<MAX_PLAYER_NUMBER;i++)
        {
            if(GetTeammatePosition(i+1,tm).GetDistanceTo(_ballPos)/PLAYER_AVARIAGE_VELOCITY <= cyclePass)
            {
                notFind = false;
                //set the action parameters
                (*act).actType = PAT_PASS;
                (*act).actRegion.top = _ballPos.GetY()+REGION_TOP_BUF;
                (*act).actRegion.bottom = _ballPos.GetY()+REGION_BOTTOM_BUF;
                (*act).actRegion.left = _ballPos.GetX()+REGION_LEFT_BUF;
                (*act).actRegion.right = _ballPos.GetX()+REGION_RIGHT_BUF;
                (*act).actMate = i+1;
                //l->Log(10,"time == %d , threshold == %d , cyclePass == %f , the pass properity == %f",GetCurrentTime(),PASS_CYCLE_MAX_THRESHOLD,cyclePass,(PASS_CYCLE_MAX_THRESHOLD-cyclePass)/PASS_CYCLE_MAX_THRESHOLD);
                return (PASS_CYCLE_MAX_THRESHOLD-cyclePass)/PASS_CYCLE_MAX_THRESHOLD;
            }
        }
    }
    
    if(notFind)
    {
        //set the action as default
        (*act).actType = PAT_PASS;
        (*act).actRegion.top = _ballPos.GetY()+REGION_TOP_BUF;
        (*act).actRegion.bottom = _ballPos.GetY()+REGION_BOTTOM_BUF;
        (*act).actRegion.left = _ballPos.GetX()+REGION_LEFT_BUF;
        (*act).actRegion.right = _ballPos.GetX()+REGION_RIGHT_BUF;
        (*act).actMate = 0;
    }

    return 0.8;    
}

double WorldModel::IsDribble(VecPos ballPos,VecVel ballVel,int uniform,int tm,PlayerActS *act)
{
    if( (*act).actMate == uniform)
    {
        (*act).actType = PAT_DRIBBLE;
        return ballVel.GetMagnitude()>=DRIBBLE_BALL_SPEED_THRESHOLD ? 
                1/( ((ballVel.GetMagnitude()-DRIBBLE_BALL_SPEED_THRESHOLD)+1)*((ballVel.GetMagnitude()-DRIBBLE_BALL_SPEED_THRESHOLD)+1) )
                :1;
    }
    else
        return 0;
}

//return the property of shoot action
double WorldModel::IsShoot(VecPos ballPos,VecVel ballVel,VecPos oppGoal,PlayerActS *act)
{
    double r = 1;
    double y;

    //ball speed must exceed the shoot ball speed threshold
    if(ballVel.GetMagnitude()<=SHOOT_BALL_SPEED_THRESHOLD)
        r *= 0;
    else
        r *= (ballVel.GetMagnitude()-SHOOT_BALL_SPEED_THRESHOLD)/(ss->ball_accel_max-SHOOT_BALL_SPEED_THRESHOLD);
    
    //ball distance to goal must be shorter than the shoot distance to goal threshold
    if(ballPos.GetDistanceTo(oppGoal)>=SHOOT_DISTANCE_TO_GOAL_THRESHOLD)
        r *= 0;
    else
        r *= (SHOOT_DISTANCE_TO_GOAL_THRESHOLD - ballPos.GetDistanceTo(oppGoal) )/SHOOT_DISTANCE_TO_GOAL_THRESHOLD;

    //ball velocity must towards the goal direction
    if(oppGoal.GetX() * ballVel.GetX()<=0)
        r *= 0;
    else
    {
        //calculate the intercept point of the ball heading ray with the opponent goal line
        y = ballVel.GetY() * ( oppGoal.GetX() - ballPos.GetX() ) / ballVel.GetX() + ballPos.GetY();
        if(fabs(y)<=SHOOT_GOAL_HALF_WIDTH_THRESHOLD)
            r *= 1;
        else
            r *= 1/ ( ( (fabs(y)-SHOOT_GOAL_HALF_WIDTH_THRESHOLD)*1000+1)*( (fabs(y)-SHOOT_GOAL_HALF_WIDTH_THRESHOLD)*1000+1) );
    }
    
    (*act).actType = PAT_SHOOT;
    return r;
}


PlayerActS WorldModel::GetTeammateIntention(int uniform,int tm)
{
    PlayerActS res;//,resS,resP,resD;
    double shootScore,passScore,dribbleScore;

    VecPos oppGoal;
    if(ourSide==FIELD_SIDE_L)
        oppGoal.SetVect(52.5,0);
    else
        oppGoal.SetVect(-52.5,0);

    VecVel ballVel = GetBallVelocity(tm);
    VecPos ballPos = GetBallPosition(tm);
    
    shootScore = IsShoot(ballPos,ballVel,oppGoal,&res);
    passScore = IsPass(ballPos,ballVel,tm,&res);
    dribbleScore = IsDribble(ballPos,ballVel,uniform,tm,&res);
/*    
    if(shootScore>=passScore && shootScore>=dribbleScore)
        res = resS;
    else if(passScore>=dribbleScore && passScore>=shootScore)
        res = resP;
    else
        res = resD;
*/
    l->Log(8,"shootScore == %f , passScore == %f , dribbleScore == %f",shootScore,passScore,dribbleScore);
    l->Log(8,"player %d's action is %s, the region is ( %f,%f ), the actMate is %d",
                uniform,
                res.actType==PAT_DRIBBLE?"dribble":(res.actType==PAT_PASS?"pass":(res.actType==PAT_SHOOT?"shoot":"none")),
                (res.actRegion.left+res.actRegion.right)/2,(res.actRegion.top+res.actRegion.bottom)/2,
                res.actMate
        );

    return res;
}


double WorldModel::GetTeamActUtility(TeamActS teamAct)
{
    double res = 1;
    int     sScore,oScore;
    
    int startTime = teamAct.time;
    int endTime = teamAct.time+1;
    while( (GetOurBallOwner(endTime)).size()==0 && (GetOppBallOwner(endTime)).size()==0 )
        endTime++;
    
    VecPos ballPosS = GetBallPosition(startTime);
    VecPos ballPosO = GetBallPosition(endTime);
    
    
    double SourceScore = 0.01*ballPosS.GetX()*FIELD_LENGTH/(ballPosS.GetX()+FIELD_LENGTH/2+1)
                            + 100*ballPosS.GetX()*(FIELD_WIDTH/2-fabs(ballPosS.GetY()))/FIELD_LENGTH;
    double ObjectScore = 0.01*ballPosO.GetX()*FIELD_LENGTH/(ballPosO.GetX()+FIELD_LENGTH/2+1)
                            + 100*ballPosO.GetX()*(FIELD_WIDTH/2-fabs(ballPosO.GetY()))/FIELD_LENGTH;

    res =  (ObjectScore  -  SourceScore)/3500;
    
    //cout<<"res == "<<res<<endl;
    
    l->Log(8,"teamAct.organizer == %d",teamAct.organizer);
    
    switch(teamAct.mateAct[teamAct.organizer-1].actType)
    {
        case PAT_POS:
            l->Log(8,"organizer is positioning");
            return 1;
            break;
        case PAT_PASS:
            l->Log(8,"organizer is passing");
            return res;
            break;
        case PAT_DRIBBLE:
            l->Log(8,"organizer is dribbling");
            return res;
            break;
        case PAT_SHOOT:
            sScore = GetOurScore(startTime);
            oScore = GetOurScore(endTime);
            l->Log(8,"organizer is shooting");
            if(oScore-sScore>0)
                return 1;
            else
                return 0;
            break;
        default:
            break;
    }
    
    return res;
}



