/***************************************************************************
                                  WorldModel.h 
                             -------------------
    begin                : 2001
    copyright            : (C)2001-2003 Multi-Agent Systems Laboratory,
                           University of Science and Technology of China 
                           Zhanxiang Huang, Yang Yang and Xiaoping Chen.
    email                : huangzx@mail.ustc.edu.cn
 ***************************************************************************/

#ifndef _WORLDMODEL_H_
#define _WORLDMODEL_H_

#include <list>
#include <vector>
#include "Geometry.h"
#include "Types.h"
#include "ServerSetting.h"

using namespace std;


class Object{
public:
        Object();
        
        // interface of Setting informations
        virtual         bool         SetPosition(const VecPos &pos);
        virtual         bool         SetPositionX( const double x);
        virtual         bool         SetPositionY(const double y);
        
        virtual         bool         SetVelocity(const VecVel &vel);
        virtual         bool         SetVelocityX(const double x);
        virtual         bool         SetVelocityY(const double y);
        
        //  interface of getting informations
        
        virtual         VecPos    GetPosition() const;
        virtual         double    GetPositionX() const;
        virtual         double    GetPositionY() const;
        
        virtual         VecVel    GetVelocity() const;
        virtual         double    GetVelocityX() const;
        virtual         double    GetVelocityY() const;
        
        virtual         double    GetSpeed() const;
        virtual         double    GetDistanceTo(const Object o);          // getting distance to object o 
    
        virtual          bool      IsDataValid() const;
        virtual          void      SetDataValidation(bool b);      // Чλ
       
protected:
        VecPos    position;       // position of the object
        VecVel    velocity;       // velocity of the object

        bool      bDataValid;     // indicate that these information Is validate
};


class BallObject : public Object{
public:
        BallObject();
private:

};


class PlayerObject : public Object{
public:
        PlayerObject();
        
        bool         SetBodyAng(const AngDeg d);
        bool         SetNeckAng(const AngDeg d);
        
        AngDeg GetBodyAng() const;
        AngDeg GetNeckAng() const;
        
private:

        AngDeg   bodyAng;
        AngDeg   neckAng;
};


// This class describes situations of the world  at cycle x
class WMSnapShot{
public:
        WMSnapShot();
        
        // interface of Setting informations
        bool    SetPlayMode(const PlayModeT mode);
        bool    SetOurScore(const int score);
        bool    SetOppScore(const int score);
        
        bool    SetBallPosition(const VecPos &pos) ;
        bool    SetBallPositionX(const double x) ;
        bool    SetBallPositionY(const double y) ;
        bool    SetBallVelocity(const VecVel &vel) ;
        bool    SetBallVelocityX(const double x) ;
        bool    SetBallVelocityY(const double y) ;

        // Set informations for teammate i
        bool    SetTeammatePosition( const VecPos &pos,  const int uniform);
        bool    SetTeammatePositionX( const double x,  const int uniform);
        bool    SetTeammatePositionY(const double y, const int uniform);
        bool    SetTeammateVelocity(const VecVel &vel,  const int uniform);
        bool    SetTeammateVelocityX(const double x,  const int uniform);
        bool    SetTeammateVelocityY(const double y,  const int uniform);
        bool    SetTeammateBodyAng(const AngDeg ang, const  int uniform);
        bool    SetTeammateNeckAng(const AngDeg ang, const  int uniform);
        
        bool    SetOpponentPosition(VecPos &pos,  const int uniform);
        bool    SetOpponentPositionX( const double x,  const int uniform);
        bool    SetOpponentPositionY(const double y, const int uniform);
        bool    SetOpponentVelocity(const VecVel &vel,  const int uniform);
        bool    SetOpponentVelocityX(const double x,  const int uniform);
        bool    SetOpponentVelocityY(const double y,  const int uniform);
        bool    SetOpponentBodyAng(const AngDeg ang,  const int uniform);
        bool    SetOpponentNeckAng(const AngDeg ang, const  int uniform);
        
        bool    AddEvent(const EventT event);
   
        bool    SetTimeStamp(const int t);
         
        //interface of getting informations
        PlayModeT       GetPlayMode() const;
        int             GetOurScore() const;
        int             GetOppScore() const;
        
        VecPos          GetBallPosition() const;
        double          GetBallPositionX() const;
        double          GetBallPositionY() const;
        VecVel          GetBallVelocity() const;
        double          GetBallVelocityX() const;
        double          GetBallVelocityY() const;
        double          GetBallSpeed() const;
        
        VecPos          GetTeammatePosition( const int uniform) const;
        double          GetTeammatePositionX( const int uniform) const;
        double          GetTeammatePositionY( const int uniform) const;
        VecVel          GetTeammateVelocity( const  int uniform) const;
        double          GetTeammateVelocityX( const int uniform) const;
        double          GetTeammateVelocityY( const int uniform) const;
        double          GetTeammateSpeed( const int uniform) const;
        AngDeg          GetTeammateBodyAng(const int uniform) const;
        AngDeg          GetTeammateNeckAng( const int uniform) const;
        
        
        VecPos          GetOpponentPosition(  const  int uniform) const;
        double          GetOpponentPositionX(  const  int uniform) const;
        double          GetOpponentPositionY(  const int uniform) const;
        VecVel          GetOpponentVelocity( const  int uniform) const;
        double          GetOpponentVelocityX( const int uniform) const;
        double          GetOpponentVelocityY(  const int uniform) const;
        double          GetOpponentSpeed( const int uniform) const;
        AngDeg          GetOpponentBodyAng(  const int uniform ) const;
        AngDeg          GetOpponentNeckAng( const int uniform) const;
        
        list<EventT>&   GetEvents();

        int             GetTimeStamp() const;
        // Interface that Setting data validation
        void            SetBallDataValidation(bool b);
        void            SetTeammateDataValidation(bool b, const int uniform);
        void            SetOpponentDataValidation(bool b, const int uniform);
        
        // WMSnopShotеϢΪfalse, playmode-->PM_ILLEGAL
        // ͬʱeventsÿ, ourScore,oppScore = 0;
        void            SetAllDataInvalidate();
        
        // Interface that getting data validation
        bool            GetBallDataValidation() const;
        bool            GetTeammateDataValidation(const int uniform) const;
        bool            GetOpponentDataValidation(const int uniform) const;
        
        // ¼
        void            clearEvent();
        
private:
        // describes the playmode at current cycle
        PlayModeT      playmode;
        
        // match informations
        int            ourScore;
        int            oppScore;
        
        // timestamp for WMSnapShot
        int            timeStamp;
        
        // record the informations of ball and players
        BallObject      ball;
        PlayerObject    teammates[MAX_PLAYER_NUMBER]; // start from 0
        PlayerObject    opponents[MAX_PLAYER_NUMBER];        
        
        
        list<EventT>  events;

};


// class of  World Model
class WorldModel{
public:
        WorldModel();
        
        void    SetLogger(Logger *log);

        // interface of Setting informations
        // tm:  representing the time cycle in the WorldModel
        
        // Set current time to tm
        bool    SetCurrentTime(const int tm);
                
        bool    SetPlayMode(const PlayModeT mode, int tm);
        bool    SetOurScore(const int score, int tm);
        bool    SetOppScore(const int score, int tm);
        
        bool    SetBallPosition(const VecPos &pos, int tm) ;
        bool    SetBallPositionX(const double x, int tm) ;
        bool    SetBallPositionY(const double y, int tm) ;
        bool    SetBallVelocity(const VecVel &vel, int tm) ;
        bool    SetBallVelocityX(const double x, int tm) ;
        bool    SetBallVelocityY(const double y, int tm) ;

        // Set informations for teammate i
        bool    SetTeammatePosition( const VecPos &pos,  const int uniform, int tm);
        bool    SetTeammatePositionX( const double x,  const int uniform, int tm);
        bool    SetTeammatePositionY(const double y, const int uniform, int tm);
        bool    SetTeammateVelocity(const VecVel &vel,  const int uniform, int tm);
        bool    SetTeammateVelocityX(const double x,  const int uniform, int tm);
        bool    SetTeammateVelocityY(const double y,  const int uniform, int tm);
        bool    SetTeammateBodyAng(const AngDeg ang, const  int uniform, int tm);
        bool    SetTeammateNeckAng(const AngDeg ang, const  int uniform, int tm);
        
        bool    SetOpponentPosition(VecPos &pos,  const int uniform, int tm);
        bool    SetOpponentPositionX( const double x,  const int uniform, int tm);
        bool    SetOpponentPositionY(const double y, const int uniform, int tm);
        bool    SetOpponentVelocity(const VecVel &vel,  const int uniform, int tm);
        bool    SetOpponentVelocityX(const double x,  const int uniform, int tm);
        bool    SetOpponentVelocityY(const double y,  const int uniform, int tm);
        bool    SetOpponentBodyAng(const AngDeg ang,  const int uniform, int tm);
        bool    SetOpponentNeckAng(const AngDeg ang, const  int uniform, int tm);
        
        bool    AddEvent(const EventT event, int tm);
              
        //interface of getting informations
        int             GetCurrentTime() const;
        
        PlayModeT       GetPlayMode(int tm) const;
        int             GetOurScore(int tm) const;
        int             GetOppScore(int tm) const;
        
        VecPos          GetBallPosition(int tm) const;
        double          GetBallPositionX(int tm) const;
        double          GetBallPositionY(int tm) const;
        VecVel          GetBallVelocity(int tm) const;
        double          GetBallVelocityX(int tm) const;
        double          GetBallVelocityY(int tm) const;
        double          GetBallSpeed(int tm) const;
        
        VecPos          GetTeammatePosition( const int uniform,int tm) const;
        double          GetTeammatePositionX( const int uniform,int tm) const;
        double          GetTeammatePositionY( const int uniform,int tm) const;
        VecVel          GetTeammateVelocity( const  int uniform,int tm) const;
        double      	GetTeammateVelocityX( const int uniform,int tm) const;
        double          GetTeammateVelocityY( const int uniform,int tm) const;
        double          GetTeammateSpeed( const int uniform,int tm) const;
        AngDeg          GetTeammateBodyAng(const int uniform,int tm) const;
        AngDeg          GetTeammateNeckAng( const int uniform,int tm) const;
        
        
        VecPos          GetOpponentPosition(  const  int uniform,int tm) const;
        double          GetOpponentPositionX(  const  int uniform,int tm) const;
        double          GetOpponentPositionY(  const int uniform,int tm) const;
        VecVel          GetOpponentVelocity( const  int uniform,int tm) const;
        double          GetOpponentVelocityX( const int uniform,int tm) const;
        double          GetOpponentVelocityY(  const int uniform,int tm) const;
        double          GetOpponentSpeed( const int uniform,int tm) const;
        AngDeg          GetOpponentBodyAng(  const int uniform ,int tm) const;
        AngDeg          GetOpponentNeckAng( const int uniform,int tm) const;

        double          GetPlayerPositionX(SideT side, const  int uniform, int tm)  const;
        double          GetPlayerPositionY(SideT side, const int uniform, int tm)  const;
        VecVel          GetPlayerVelocity(SideT side, const  int uniform, int tm)  const;
        double          GetOpponentVelocityX(SideT side, const int uniform, int tm)  const;
        double          GetOpponentVelocityY(SideT side,  const int uniform, int tm)  const;
        double          GetOpponentSpeed(SideT side, const int uniform, int tm)  const;
        AngDeg          GetOpponentBodyAng(SideT side,  const int uniform, int tm )  const;
        AngDeg          GetOpponentNeckAng(SideT side, const int uniform, int tm)  const;

        list<EventT>&   GetEvents(int tm);
        WMSnapShot      GetWMSnapShot(const int tm) const;
                
        // Update functions
        
        // úʹǰ״̬Ϊwms,ͬʱѹvector field
        bool    UpdateWorldModel( WMSnapShot  &wms);

        bool    IsOurGoalieNumSet() const;
        bool    IsOppGoalieNumSet() const;
        // ԱºţͬʱbGoalieUniformSetΪ true
        void    SetOurGoalieNum(const int num);
        void    SetOppGoalieNum(const int num);
        int     GetOurGoalieNum();
        int     GetOppGoalieNum();
        
        bool    IsOurTeamNameSet() const;
        bool    IsOppTeamNameSet() const;
        // ͬʱbTeamNameSetΪtrue
        void    SetOurTeamName(string &teamName);
        void    SetOppTeamName(string &teamName);
        
        // Set our side
        void SetOurSide(SideT side);
        void SetOppSide(SideT side);
        SideT GetOurSide();
        SideT GetOppSide();
        
        void SetServerParameters(ServerSetting* css, PlayerSetting* cps, PlayerTypeSetting* cpts);
        
        // judging functions
        bool        IsBallKickableForTeammate(int uniform, int tm);
        bool        IsBallKickableForOpponent(int uniform, int tm);
        bool        IsBallKickableForPlayer(SideT side, int uniform, int tm);
        
        double      IsPass(VecPos ballPos,VecVel ballVel,int tm,PlayerActS *act);
        double      IsDribble(VecPos ballPos,VecVel ballVel,int uniform,int tm,PlayerActS *act);
        double      IsShoot(VecPos ballPos,VecVel ballVel,VecPos oppGoal,PlayerActS *act);
        
        PlayerActS  GetTeammateIntention(int uniform , int tm);
        double      GetTeamActUtility(TeamActS teamAct);
        
        // get hign level information
        vector<int>   GetOurBallOwner(int tm);
        vector<int>   GetOppBallOwner(int tm);
        void        OutBallOwner(vector<int> ballOwner);
                
private:
        vector<WMSnapShot>    field;   // informations in the field
        int                   ctm;     // current time in the match
        
        //  information about the teams
        string     oppTeamName;
        bool       bOppTeamNameSet;
        string     ourTeamName;
        bool       bOurTeamNameSet;
        
        int        oppGoalieUniform;              // uniform num of goalies
        bool       bOppGoalieUniformSet;          // indicated whether goalie uniform have Set
        int        ourGoalieUniform;
        bool       bOurGoalieUniformSet;            
        
        SideT      ourSide;
        SideT      oppSide;

        ServerSetting *ss;
        PlayerSetting *ps;
        PlayerTypeSetting *pts;       
        Logger *l;
};


#endif

